from os import path

import pytest

tools = pytest.importorskip("niftypet.ninst.tools")


def test_resources():
    assert path.exists(tools.path_resources)
    assert tools.resources.DIRTOOLS


def test_gpu():
    assert tools.dev_info == tools.gpuinfo



"""Resources file for NiftyPET NIPET and NIMPA etc."""
__author__ = ("Pawel J. Markiewicz", "Casper O. da Costa-Luis")
__copyright__ = "Copyright 2018-20"
from math import ceil, pi

try:
    from numpy import array
except ImportError:

    def array(x):
        return x


# > logging represented by an integer: 10, 20, 30... for DEBUG, INFO, WARNING...
# > as it is in Python package logging, which is also used here.
LOG = 20

# Hardware (CT-based) mu-maps, which come with the mMR scanner.
# The names may be different
hrdwr_mu = [
    "umap_HNMCL_10606489.v.hdr",  # (1) Head and neck lower coil
    "umap_HNMCU_10606489.v.hdr",  # (2) Head and neck upper coil
    "umap_SPMC_10606491.v.hdr",  # (3) Spine coil
    "umap_PT_2291734.v.hdr",  # (4) Table
]

# Radioisotope look-up table
riLUT = {
    "Ge68": {"BF": 0.891, "thalf": 270.9516 * 24 * 60 * 60},
    "Ga68": {"BF": 0.891, "thalf": 67.71 * 60},
    "F18": {"BF": 0.967, "thalf": 109.77120 * 60},
    "C11": {"BF": 0.998, "thalf": 20.38 * 60},
    "O15": {"BF": 0.999, "thalf": 122.2416},
}

# -----------------------------------------------------
# The name and path to the NiftyPET tools (software)
DIRTOOLS = "NiftyPET_tools"
MSVC_VRSN = "Visual Studio 12 2013 Win64"
CMAKE_TLS_PAR = ""  # -DUSE_SSE=OFF'
# PATHTOOLS = os.path.join('/chosen/path/', DIRTOOLS)
# > path to Python wrapper of Vinci
VINCIPATH = ""

# > path to reference images for testing NiftyPET
REFPATH = ""
# -----------------------------------------------------

# -----------------------------------------------------
# DO NOT MODIFY BELOW--DONE AUTOMATICALLY
# # # start GPU properties # # #
# # # end GPU properties # # #

# paths to apps and tools needed by NiftyPET
# # # start NiftyPET tools # # #
# # # end NiftyPET tools # # #
# -----------------------------------------------------

# enable xnat module
ENBLXNAT = False
# enable Agg
ENBLAGG = False
# compile DCM2NIIX, otherwise download a compiled version for the system used
CMPL_DCM2NIIX = False

# ============ SIEMENS mMR SCANNER C O N S T A N T S ===============

# > LM header offset in bytes (for mMR it is in a separate DICOM format)
LMOFF = 0

# > bytes per event in list mode data
BPE = 4

# number of rings (axially) and crystals (transaxially)
NRNG = 64

# number of crystals transaxially
NCRS = 504

# > idealised crystal surface (used for scatter modelling)
SRFCRS = 0.1695112

# reduced number of crystals by the gaps (dead crystals)
NCRSR = 448

# maximum ring difference
MRD = 60

# number of linear indexes for 1/2 Michelogram (NRNG**2/2 + NRNG/2)
NLI2R = 2080 - 6

# number of angular indexes in a 2D sinogram
A = 252
# number of bin indexes in a 2D sino
W = 344
H = W / 2

NSN11 = 837
NSN1 = 4084
NSN64 = NRNG * NRNG
# 0: SSRB, 1: span-1, or 11: span-11 (11, default)
SPAN = 11

RNG_STRT = 0
RNG_END = 64

# no of sinos in a segment out of 11 segments
seg = array([127, 115, 115, 93, 93, 71, 71, 49, 49, 27, 27])

# minimum and maximum ring difference for each segment
minrd = array([-5, -16, 6, -27, 17, -38, 28, -49, 39, -60, 50])
maxrd = array([5, -6, 16, -17, 27, -28, 38, -39, 49, -50, 60])
# ----------


# ------------------------------------------------------
# > transaxial projection parameters (should be in
# > with the parameters as defined in def.h for C files)

# > parameters for each transaxial LOR
NTT = 10

# > all voxels intersected by a given LOR
NTV = 907
# ------------------------------------------------------

# number of direct sinograms (i.e., for segment 0)
SEG0 = 127

# Reference image size (usually the default from Siemens)
# and GPU dimensions for optimal execution
# ~~~
SO_IMZ = 127
SO_IMY = 344
SO_IMX = 344
SO_VXX = 0.208626
SO_VXZ = 0.203125
SO_VXY = SO_VXX

SZ_IMZ = 127
SZ_IMY = 320
SZ_IMX = 320
SZ_VOXY = 0.208626
SZ_VOXZ = 0.203125
# ~~~
# SO_IMZ = 127
# SO_IMY = 384
# SO_IMX = 384
# SO_VXX = 0.1669
# SO_VXZ = 0.203125
# SO_VXY = SO_VXX

# SZ_IMZ = 127
# SZ_IMY = 384
# SZ_IMX = 384
# SZ_VOXY = 0.1669
# SZ_VOXZ = 0.203125
# ~~~
SIGMA_RM = 0

# > radius PSF kernel size used in CUDA convolution
RSZ_PSF_KRNL = 8

# ~~~
# inverse size
SZ_VOXZi = round(1 / SZ_VOXZ, 6)
# squared radius of the transaxial field of view
TFOV2 = 890.0

# -------Scatter image size in x,y,z directions
# target scale factors for scatter mu-map and emission image respectively
# transmission (mu-map)
TRGTSCT = [0.5, 0.33]
SS_IMX = int(ceil(TRGTSCT[0] * SO_IMX) // 2 * 2)
SS_IMY = int(ceil(TRGTSCT[0] * SO_IMY) // 2 * 2)
SS_IMZ = int(ceil(TRGTSCT[0] * SO_IMZ) // 2 * 2 - 1)
SS_VXY = round((SO_VXY * SO_IMX) / SS_IMX, 6)
SS_VXZ = round((SO_VXZ * SO_IMZ) / SS_IMZ, 6)
IS_VXZ = round(1 / SS_VXZ, 6)
# scaling [z,y,x]
SCTSCLMU = [float(SS_IMZ) / SO_IMZ, float(SS_IMY) / SO_IMY, float(SS_IMX) / SO_IMX]
# emission
SSE_IMX = int(ceil(TRGTSCT[1] * SO_IMX) // 2 * 2)
SSE_IMY = int(ceil(TRGTSCT[1] * SO_IMY) // 2 * 2)
SSE_IMZ = int(ceil(TRGTSCT[1] * SO_IMZ) // 2 * 2 + 1)
SSE_VXY = round((SO_VXY * SO_IMX) / SSE_IMX, 6)
SSE_VXZ = round((SO_VXZ * SO_IMZ) / SSE_IMZ, 6)
# scaling [z,y,x]
SCTSCLEM = [float(SSE_IMZ) / SO_IMZ, float(SSE_IMY) / SO_IMY, float(SSE_IMX) / SO_IMX]


# # scaling for the emission image [z,y,x]
# SCTSCLEM = [0.34, 0.33, 0.33]
# # scaling for the mu-map
# SCTSCLMU = [0.499, 0.5, 0.5]

# SS_IMX = int(ceil(SCTSCLMU[2]*SO_IMX)//2*2)#172
# SS_IMY = int(ceil(SCTSCLMU[1]*SO_IMY)//2*2)#172
# SS_IMZ = int(ceil(SCTSCLMU[0]*SO_IMZ)//2*2-1)#63
# SS_VXY = round((SO_VXY*SO_IMX)/SS_IMX,6) # 0.417252 #
# SS_VXZ = round((SO_VXZ*SO_IMZ)/SS_IMZ,6) # 0.409474 #
# IS_VXZ = round(1/SS_VXZ,6)

# SSE_IMX = int(ceil(SCTSCLEM[2]*SO_IMX)//2*2) #114
# SSE_IMY = int(ceil(SCTSCLEM[1]*SO_IMY)//2*2) #114
# SSE_IMZ = int(ceil(SCTSCLEM[0]*SO_IMZ)//2*2-1) #43

# SSE_VXY = round((SO_VXY*SO_IMX)/SSE_IMX,6) #0.629538
# SSE_VXZ = round((SO_VXZ*SO_IMZ)/SSE_IMZ,6) #0.599927
# -------

# > decay correction
DCYCRR = True

# --- Time of Flight ---
# speed of light
CLGHT = 29979245800  # cm/s
# coincidence time window [ps]
CWND = 5859.38e-12

# number of TOF bins
TOFBINN = 1
# size of TOF bin in [ps]
TOFBINS = 390e-12
# size of TOF BIN in cm of travelled distance
TOFBIND = TOFBINS * CLGHT
# inverse of the above
ITOFBIND = 1 / TOFBIND


# ring radius
R = 32.8
# effective ring radius accounting for the depth of interaction
RE = R + 0.67  # 0.67

RE_2 = float("{0:.6f}".format(RE ** 2))

# > inverse of the radius
IRE = float("{0:.6f}".format(RE ** -1))

# axial crystal width
AXR = 0.40625

# crystal angle
ALPHA = 0.714286 * pi / 180  # 2*pi/NCRS

# crystal gap period
TGAP = 9

# crystal gap offset (used for getting the sino gaps right at the position)
OFFGAP = 1

# --- FOR SCATTER ---
# electron radius **2
R02 = 7.940787449825884e-26
# detection lower energy threashold
LLD = 430000
E511 = 511008
# energy resolution
ER = 0  # 0.154
# discretisation of the scatter angle spectrum
NCOS = 256
# cosine of maximum allowed scatter angle
COSUPSMX = 0.725  # 0.58 #0.722 #Elow = E511/(2-cos(upsmx))
# step of the discreatisation
COSSTP = (1 - COSUPSMX) / (NCOS - 1)
# inverse of the step
ICOSSTP = 1 / COSSTP

# intensity percentage threshold of voxels to be considered in the image
ETHRLD = 0.05


# =================================================================================================
def get_gpu_constants(Cnt=None):
    """Return a dictionary of GPU related constants"""
    if Cnt is None:
        Cnt = {}

    for k in [
        "DEV_ID",  # device id; used for choosing the GPU device for calculations
        "CC_ARCH",  # chosen device architectures for NVCC compilation
    ]:
        val = globals().get(k)
        if val is not None:
            Cnt[k.replace("_", "")] = val

    return Cnt


# =================================================================================================
def get_setup(Cnt=None):
    """Return a dictionary of GPU, mu-map hardware and third party set-up."""
    if Cnt is None:
        Cnt = {}

    # the name of the folder for NiftyPET tools
    Cnt["DIRTOOLS"] = DIRTOOLS

    # additional paramteres for compiling tools with cmake
    Cnt["CMAKE_TLS_PAR"] = CMAKE_TLS_PAR

    # hardware mu-maps
    Cnt["HMULIST"] = hrdwr_mu

    # Microsoft Visual Studio Compiler version
    Cnt["MSVC_VRSN"] = MSVC_VRSN

    # GPU related setup
    Cnt = get_gpu_constants(Cnt)

    for k in [
        "PATHTOOLS",
        "RESPATH",  # image processing setup
        "REGPATH",
        "DCM2NIIX",
        "HMUDIR",  # hardware mu-maps
        "VINCIPATH",
        "REFPATH",  # > testing
    ]:
        val = globals().get(k)
        if val is not None:
            Cnt[k] = val

    Cnt["ENBLXNAT"] = ENBLXNAT
    Cnt["ENBLAGG"] = ENBLAGG
    Cnt["CMPL_DCM2NIIX"] = CMPL_DCM2NIIX

    return Cnt


# =================================================================================================
def get_mmr_constants():
    """
    Put all the constants together in a dictionary
    """

    Cnt = {
        "LOG": LOG,
        "VERBOSE": False,
        "BPE": BPE,
        "LMOFF": LMOFF,
        "ISOTOPE": "F18",
        "DCYCRR": DCYCRR,
        "ALPHA": ALPHA,
        "NRNG": NRNG,
        "NCRS": NCRS,
        "NCRSR": NCRSR,
        "SRFCRS": SRFCRS,
        "NBCKT": 224,
        "NSANGLES": A,
        "NSBINS": W,
        "Naw": -1,  # number of total active bins per 2D sino
        "NSN11": NSN11,  # number of sinos in span-11
        "NSN1": NSN1,  # number of sinos in span-1
        "NSN64": NSN64,  # number of sinos in span-1 with no MRD limit
        "MRD": MRD,  # maximum ring difference RD
        "SPN": SPAN,  # span-1 (1), span-11 (11), ssrb (0)
        "TFOV2": TFOV2,  # squared radius of TFOV
        "RNG_STRT": RNG_STRT,  # limit axial extension by defining start and end ring
        "RNG_END": RNG_END,  # only works with span-1 (Cnt['SPN']==1)
        "SS_IMZ": SS_IMZ,  # Scatter mu-map image size
        "SS_IMY": SS_IMY,
        "SS_IMX": SS_IMX,
        "SS_VXZ": SS_VXZ,
        "SS_VXY": SS_VXY,
        "IS_VXZ": IS_VXZ,
        "SSE_IMZ": SSE_IMZ,  # Scatter emission image size
        "SSE_IMY": SSE_IMY,
        "SSE_IMX": SSE_IMX,
        "SSE_VXZ": SSE_VXZ,
        "SSE_VXY": SSE_VXY,
        "SZ_IMZ": SZ_IMZ,  # GPU optimised image size
        "SZ_IMY": SZ_IMY,
        "SZ_IMX": SZ_IMX,
        "SZ_VOXZ": SZ_VOXZ,
        "SZ_VOXY": SZ_VOXY,
        "SZ_VOXZi": SZ_VOXZi,
        "SO_IMZ": SO_IMZ,  # Original image size (from Siemens)
        "SO_IMY": SO_IMY,
        "SO_IMX": SO_IMX,
        "SO_VXZ": SO_VXZ,
        "SO_VXY": SO_VXY,
        "SO_VXX": SO_VXX,
        "SIGMA_RM": SIGMA_RM,  # resolution modelling sigma
        "RSZ_PSF_KRNL": RSZ_PSF_KRNL,  # radius PSF kernel size used in CUDA convolution
        "NTT": NTT,
        "NTV": NTV,
        "NSEG0": SEG0,
        "R_RING": RE,  # effective ring radius
        "R_2": RE_2,
        "IR_RING": IRE,
        "R": R,
        "SEG": seg,
        "MNRD": minrd,
        "MXRD": maxrd,
        # > scatter moved to scatter LUTs script in sct folder
        # 'NSRNG':NSRNG, # number of scatter rings for modelling
        # 'SCTRNG':sct_irng, # scatter ring indexes
        "TGAP": TGAP,
        "OFFGAP": OFFGAP,
        "AXR": AXR,
        "R02": R02,  # squared electron radius
        "LLD": LLD,  # lower energy threashold
        "E511": E511,
        "ER": ER,  # energy resolution
        "COSUPSMX": COSUPSMX,  # cosine of max allowed scatter angle
        "NCOS": NCOS,  # number of cos samples for LUT
        "COSSTP": COSSTP,  # cosine step
        "ICOSSTP": ICOSSTP,  # inverse of cosine step
        "ETHRLD": ETHRLD,  # intensity emission image threshold (for scatter modelling)
        "CLGHT": CLGHT,  # speed of light [cm/s]
        "CWND": CWND,  # coincidence time window [ps]
        "TOFBINN": TOFBINN,  # number of TOF bins
        "TOFBINS": TOFBINS,  # TOF bin width [ps]
        "TOFBIND": TOFBIND,
        "ITOFBIND": ITOFBIND,
        # affine and image size for the reconstructed image,
        # assuming the centre of voxels in mm
        "AFFINE": array(
            [
                [-10 * SO_VXX, 0.0, 0.0, 5.0 * SO_IMX * SO_VXX],  # +5.*SO_VXX
                [0.0, 10 * SO_VXY, 0.0, -5.0 * SO_IMY * SO_VXY],  # +5.*SO_VXY
                [0.0, 0.0, 10 * SO_VXZ, -5.0 * SO_IMZ * SO_VXZ],  # -5.*SO_VXZ
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
        "IMSIZE": array([SO_IMZ, SO_IMY, SO_IMX]),
        "BTP": 0,  # 1:non parametric bootstrap, 2: parametric bootstrap (recommended)
        "BTPRT": 1.0,  # Ratio of bootstrapped/original events (enables downsampling)
        "SCTSCLEM": SCTSCLEM,
        "SCTSCLMU": SCTSCLMU,
    }

    # get the setup for GPU and third party apps
    Cnt = get_setup(Cnt=Cnt)

    return Cnt



from io import StringIO

from pytest import raises

from niftypet.ninst import install_tools as tls


def test_query(capsys, monkeypatch):
    monkeypatch.setattr("sys.stdin", StringIO("y"))
    assert tls.query_yesno("hello")
    out, _ = capsys.readouterr()
    assert "hello [Y/n]" in out
    assert "Please respond with" not in out

    monkeypatch.setattr("sys.stdin", StringIO("N"))
    assert not tls.query_yesno("hello")
    out, _ = capsys.readouterr()
    assert "Please respond with" not in out

    monkeypatch.setattr("sys.stdin", StringIO("what\ny"))
    assert tls.query_yesno("hello")
    out, _ = capsys.readouterr()
    assert "Please respond with" in out


def test_check_platform():
    tls.check_platform()


def test_check_depends():
    deps = tls.check_depends()
    assert not {"cmake", "ninja", "cuda", "git"} - deps.keys()
    assert any(isinstance(i, tuple) for i in deps.values())


def test_check_version():
    deps = tls.check_version({})
    assert not {"RESPATH", "REGPATH", "DCM2NIIX", "HMUDIR"} - deps.keys()


def test_install_tool(tmp_path, monkeypatch):
    dname = tmp_path / "install_tool"
    monkeypatch.setenv("PATHTOOLS", str(dname))
    Cnt = {"DIRTOOLS": "NiftyPET_tools_test"}
    assert not dname.exists()
    with raises(ValueError):
        tls.install_tool("", Cnt)
    assert (dname / Cnt["DIRTOOLS"]).is_dir()



import os

import pytest


@pytest.fixture(scope="session", autouse=True)
def noninteractive():
    if os.getenv("DISPLAY", False):
        pytest.skip("Need to call pytest with DISPLAY=''")


@pytest.fixture(scope="session")
def nvml():
    import pynvml

    try:
        pynvml.nvmlInit()
    except pynvml.NVMLError as exc:
        pytest.skip(str(exc))
    return pynvml



import logging

from niftypet.ninst import cudasetup as cs


def test_find_cuda(caplog):
    with caplog.at_level(logging.WARNING):
        cuda = cs.find_cuda()
    assert cuda or caplog.record_tuples


def test_dev_setup(caplog, nvml):
    with caplog.at_level(logging.WARNING):
        ccstr = cs.dev_setup()
    assert ccstr or caplog.record_tuples


def test_resources_setup(caplog):
    with caplog.at_level(logging.WARNING):
        gpu = cs.resources_setup(gpu=False)
    assert not caplog.record_tuples
    assert not gpu



from niftypet.ninst.dinf import dev_info


def test_dev_info(capsys, nvml):
    devs = dev_info()
    out, err = capsys.readouterr()
    assert not any((out, err))
    dev_info(showprop=True)
    out, err = capsys.readouterr()
    assert not err
    assert not devs or out



"""
Install tools for NiftyPET including:
* NiftyReg
* dcm2niix
"""
import logging
import multiprocessing
import os
import platform
import re
import sys
from functools import wraps
from os import fspath, path
from pathlib import Path
from shutil import rmtree
from subprocess import CalledProcessError, check_output, run
from textwrap import dedent

from miutil.fdio import extractall
from miutil.web import urlopen_cached

from . import cudasetup as cs

if os.getenv("DISPLAY", False):
    from tkinter import Tk
    from tkinter.filedialog import askdirectory as ask

    @wraps(ask)
    def askdir(title, initialdir):
        Tk().withdraw()
        res = ask(title=title, initialdir=initialdir)
        Tk().destroy()
        return res


else:

    def askdir(title, initialdir):
        res = input(title + (f" [{initialdir}]: " if initialdir else ": "))
        return initialdir if res == "" else res


log = logging.getLogger(__name__)

# NiftyReg
repo_reg = "https://github.com/NiftyPET/niftyreg"
# repo_reg = 'https://cmiclab.cs.ucl.ac.uk/mmodat/niftyreg.git'
# 'git://git.code.sf.net/p/niftyreg/git'
sha1_reg = "gcc9-omp-fix"
# 'f673b7837c0824f55dedb1534b32b55bf68a2823'
# '6bf84b492050a4b9a93431209babeab9bc8f14da'
# '62af1ca6777379316669b6934889c19863eaa708'
reg_ver = "1.5.68"

# dcm2niix
repo_dcm = "https://github.com/rordenlab/dcm2niix"
http_dcm_lin = repo_dcm + "/releases/download/v1.0.20200331/dcm2niix_lnx.zip"
# http_dcm_lin = repo_dcm + '/releases/download/v1.0.20190902/dcm2niix_lnx.zip'
http_dcm_win = repo_dcm + "/releases/download/v1.0.20200331/dcm2niix_win.zip"
# repo_dcm + '/releases/download/v1.0.20190902/dcm2niix_win.zip'
http_dcm_mac = repo_dcm + "/releases/download/v1.0.20200331/dcm2niix_mac.zip"
# repo_dcm + '/releases/download/v1.0.20190902/dcm2niix_mac.zip'
sha1_dcm = "485c387c93bbca3b29b93403dfde211c4bc39af6"
# 'f54be46667fce7994d2062e2623d12253c1bd968'
dcm_ver = "v1.0.20200331"  # 'v1.0.20190902'
# PREVIOUS WORKING:
# http_dcm_lin =  repo_dcm + '/releases/download/v1.0.20180622/'
# 'dcm2niix_27-Jun-2018_lnx.zip'
# http_dcm_win = repo_dcm + '/releases/download/v1.0.20180622/'
# 'dcm2niix_27-Jun-2018_win.zip'
# sha1_dcm =  '4b641113273d86ad73123816993092fc643ac62f'
# dcm_ver = '1.0.20180622'
http_dcm = {"Windows": http_dcm_win, "Linux": http_dcm_lin, "Darwin": http_dcm_mac}

# source and build folder names
dirsrc = "_src"
dirbld = "_bld"

# number of threads
ncpu = multiprocessing.cpu_count()

LOG_FORMAT = "%(levelname)s:%(asctime)s:%(name)s:%(funcName)s\n> %(message)s"


class LogHandler(logging.StreamHandler):
    """Custom formatting"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fmt = logging.Formatter(LOG_FORMAT)
        self.setFormatter(fmt)


def askdirectory(title="Folder", initialdir="~", name=""):
    """
    Args:
      initialdir (str):  default: "~"
    Returns (str):
      one of (decreasing Precedence):
      - `os.getenv(name)`
      - `input()` or `tkinter.filedialog.askdirectory()`
      - `initialdir`
    """
    initialdir = path.expanduser(initialdir)
    res = os.getenv(name, None)
    return askdir(title, initialdir) if res is None else res


def query_yesno(question):
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    prompt = " [Y/n]: "
    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()

        if choice == "":
            return True
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def check_platform():
    if not platform.system() in ["Windows", "Darwin", "Linux"]:
        log.error(
            dedent(
                f"""\
                the operating system is not supported: {platform.system()}
                only Linux, Windows and macOS are supported."""
            )
        )
        raise SystemError("unknown operating system (OS).")


def int_or_str(x):
    try:
        return int(x)
    except ValueError:
        return x


def get_version(bin, required=1, errmsg=""):
    try:
        ver = check_output([bin, "--version"]).decode("U8")
    except (CalledProcessError, FileNotFoundError):
        if required > 1:
            log.error(errmsg or f"{bin} not found")
        if required > 2:
            raise
        return False
    else:
        try:
            ver = re.findall(r"\d+\.\d[-\d.\w]*", ver, flags=re.M)[-1]
        except Exception:
            log.warn(f"could not parse {bin} --version:\n{ver}")
            return True
        return tuple(map(int_or_str, re.split("[.-]", ver)))


def check_depends(git=1, cuda=1, cmake=1, ninja=1, **kwargs):
    """
    Args:
      **kwargs: values(int):
        0: skip check
        1: check, no warn if not found
        2: required, warn if not found
        3: required, error if not found

    Returns:
      kwargs: values(truthy):
        False: not found
        True: found
        tuple: version numbers
    """
    outdct = {
        k: v
        for k, v in {
            "git": git,
            "cuda": cuda,
            "cmake": cmake,
            "ninja": ninja,
            **kwargs,
        }.items()
        if v
    }
    log.info(f"checking if [{'], ['.join(outdct)}] are installed...")

    if "cuda" in outdct:
        outdct["cuda"] = get_version(
            "nvcc", outdct["cuda"], "CUDA (nvcc) does not seem to be installed!"
        )
    if "nvcc" in outdct:
        outdct["nvcc"] = get_version(
            "nvcc", outdct["nvcc"], "CUDA (nvcc) does not seem to be installed!"
        )
    if "git" in outdct:
        outdct["git"] = get_version(
            "git",
            outdct["git"],
            (
                "git does not seem to be installed;"
                " get it from: https://git-scm.com/download"
            ),
        )
    if "cmake" in outdct:
        outdct["cmake"] = get_version(
            "cmake",
            outdct["cmake"],
            ("cmake does not seem to be installed; try conda/pip install cmake"),
        )
    if "ninja" in outdct:
        outdct["ninja"] = get_version(
            "ninja",
            outdct["ninja"],
            ("ninja does not seem to be installed; try conda/pip install ninja"),
        )

    for bin in set(outdct) - {"cuda", "nvcc", "git", "cmake", "ninja"}:
        # for p in os.getenv("PATH").split(path.pathsep):
        #     if path.isfile(path.join(p, bin)):
        #         outdct[bin] = True
        #         break
        # else:
        #     outdct[bin] = False
        outdct[bin] = get_version(bin, outdct[bin])

    return outdct


def check_version(Cnt, chcklst=None):
    """
    Check version and existence of all third-party software and input data.
    Output a dictionary with bool type of the requested bits in 'chcklst'.

    Args:
      chcklst (list): default: ["RESPATH", "REGPATH", "DCM2NIIX", "HMUDIR"]
    """
    if chcklst is None:
        chcklst = ["RESPATH", "REGPATH", "DCM2NIIX", "HMUDIR"]

    # at start, assume that nothing is present yet
    output = {}
    for itm in chcklst:
        output[itm] = False

    # niftyreg reg_resample first
    if "RESPATH" in chcklst and "RESPATH" in Cnt:
        try:
            out = check_output([Cnt["RESPATH"], "--version"]).decode("U8")
            if reg_ver in out:
                output["RESPATH"] = True
        except (CalledProcessError, FileNotFoundError):
            log.error("NiftyReg (reg_resample) either is NOT installed or is corrupt.")

    # niftyreg reg_aladin
    if "REGPATH" in chcklst and "REGPATH" in Cnt:
        try:
            out = check_output([Cnt["REGPATH"], "--version"]).decode("U8")
            if reg_ver in out:
                output["REGPATH"] = True
        except (CalledProcessError, FileNotFoundError):
            log.error("NiftyReg (reg_aladin) either is NOT installed or is corrupt.")

    # dcm2niix
    if "DCM2NIIX" in chcklst and "DCM2NIIX" in Cnt:
        try:
            out = check_output([Cnt["DCM2NIIX"], "-h"]).decode("U8")
            ver_str = re.search(r"(?<=dcm2niiX version v)\d{1,2}.\d{1,2}.\d*", out)
            if ver_str and dcm_ver in ver_str.group(0):
                output["DCM2NIIX"] = True
        except (CalledProcessError, FileNotFoundError):
            log.error("dcm2niix either is NOT installed or is corrupt.")

    # hdw mu-map list
    if "HMUDIR" in chcklst and "HMUDIR" in Cnt:
        for hi in Cnt["HMULIST"]:
            if path.isfile(path.join(Cnt["HMUDIR"], hi)):
                output["HMUDIR"] = True
            else:
                output["HMUDIR"] = False
                break

    return output


def download_dcm2niix(Cnt, dest):
    log.info(
        dedent(
            """\
        ==============================================================
        dcm2niix will be installed directly from:
        https://github.com/rordenlab/dcm2niix/releases
        =============================================================="""
        )
    )

    # -create the installation folder
    dest.mkdir(exist_ok=True)
    binpath = dest / "bin"
    binpath.mkdir(exist_ok=True)

    with urlopen_cached(http_dcm[platform.system()], dest) as fd:
        extractall(fd, binpath)

    Cnt["DCM2NIIX"] = fspath(next(binpath.glob("dcm2niix*")))
    # ensure the permissions are given to the executable
    os.chmod(Cnt["DCM2NIIX"], 755)
    # update the resources.py file in ~/.niftypet
    Cnt = update_resources(Cnt)
    return Cnt


def install_tool(app, Cnt):
    """
    Install the requested software from the git 'repo'
    and check out the version given by 'sha1'.
    """
    # pick the target installation folder for tools
    if Cnt.get("PATHTOOLS", None):
        path_tools = Path(Cnt["PATHTOOLS"])
        path_tools.mkdir(exist_ok=True)
    else:
        path_tools = Path(
            askdirectory(title="Path to place NiftyPET tools", name="PATHTOOLS")
        )
        path_tools.mkdir(exist_ok=True)
        if path_tools.name != Cnt["DIRTOOLS"]:
            path_tools /= Cnt["DIRTOOLS"]
        Cnt["PATHTOOLS"] = fspath(path_tools)
    if platform.system() not in {"Linux", "Windows"}:
        raise SystemError(
            dedent(
                """\
            =============================================================
            Only Linux and Windows operating systems are supported
            for installation of additional tools.
            =============================================================
            """
            )
        )

    # create the main tools folder
    path_tools.mkdir(exist_ok=True)
    # identify the specific path for the requested app
    if app == "niftyreg":
        repo = repo_reg
        sha1 = sha1_reg
        dest = path_tools.resolve() / "niftyreg"
    elif app == "dcm2niix":
        repo = repo_dcm
        sha1 = sha1_dcm
        dest = path_tools.resolve() / "dcm2niix"

        if not Cnt["CMPL_DCM2NIIX"]:
            # avoid installing from source, instead download the full version:
            Cnt = download_dcm2niix(Cnt, dest)
            return Cnt
    else:
        raise ValueError(f"unknown tool:{app}")

    # Check if the source folder exists and delete it, if it does
    if dest.is_dir():
        rmtree(fspath(dest))
    dest.mkdir()

    # clone the git repository
    run(["git", "clone", repo, fspath(dest / dirsrc)])
    log.info("checking out the specific git version of the software...")
    run(["git", "-C", fspath(dest / dirsrc), "checkout", sha1])

    (dest / dirbld).mkdir(exist_ok=True)
    # run cmake with arguments
    if platform.system() == "Windows":
        run(
            [
                "cmake",
                fspath(dest / dirsrc),
                "-DBUILD_ALL_DEP=ON",
                f"-DCMAKE_INSTALL_PREFIX={dest}",
                "-G",
                Cnt["MSVC_VRSN"],
            ],
            cwd=fspath(dest / dirbld),
        )
        run(
            [
                "cmake",
                "--build",
                "./",
                "-j",
                "8",
                "--config",
                "Release",
                "--target",
                "install",
            ],
            cwd=fspath(dest / dirbld),
        )
    elif platform.system() in ["Linux", "Darwin"]:
        cmd = [
            "cmake",
            fspath(dest / dirsrc),
            "-DBUILD_ALL_DEP=ON",
            f"-DCMAKE_INSTALL_PREFIX={dest}",
        ]
        if Cnt["CMAKE_TLS_PAR"] != "":
            cmd.append(Cnt["CMAKE_TLS_PAR"])
        run(cmd, cwd=fspath(dest / dirbld))
        run(
            [
                "cmake",
                "--build",
                "./",
                "-j",
                "8",
                "--config",
                "Release",
                "--target",
                "install",
                "--",
                "-j",
                str(ncpu),
            ],
            cwd=fspath(dest / dirbld),
        )

    if app == "niftyreg":
        try:
            Cnt["RESPATH"] = fspath(next((dest / "bin").glob("reg_resample*")))
            Cnt["REGPATH"] = fspath(next((dest / "bin").glob("reg_aladin*")))
        except StopIteration:
            log.error("NiftyReg has NOT been successfully installed.")
            raise SystemError("Failed Installation (NiftyReg)")
        # updated the file resources.py
        Cnt = update_resources(Cnt)
        # check the installation:
        chck_niftyreg = check_version(Cnt, chcklst=["RESPATH", "REGPATH"])
        if not all(chck_niftyreg.values()):
            log.error("NiftyReg has NOT been successfully installed.")
            raise SystemError("Failed Installation (NiftyReg)")
    elif app == "dcm2niix":
        try:
            Cnt["DCM2NIIX"] = fspath(next((dest / "bin").glob("dcm2niix*")))
        except StopIteration:
            log.error("dcm2niix has NOT been successfully installed.")
            Cnt = download_dcm2niix(Cnt, dest)
        # check the installation:
        if not check_version(Cnt, chcklst=["DCM2NIIX"]):
            log.error("dcm2niix has NOT been successfully compiled from github.")
            Cnt = download_dcm2niix(Cnt, dest)
    return Cnt


def update_resources(Cnt):
    """Update resources.py with the paths to the new installed apps."""
    # list of path names which will be saved
    key_list = ["PATHTOOLS", "RESPATH", "REGPATH", "DCM2NIIX", "HMUDIR"]

    # get the local path to NiftyPET resources.py
    path_resources = cs.path_niftypet_local()
    resources_file = path.join(path_resources, "resources.py")

    # update resources.py
    if path.isfile(resources_file):
        f = open(resources_file, "r")
        rsrc = f.read()
        f.close()
        # get the region of keeping in synch with Python
        i0 = rsrc.find("# # # start NiftyPET tools # # #")
        i1 = rsrc.find("# # # end NiftyPET tools # # #")
        pth_list = []
        for k in key_list:
            if k in Cnt:
                pth_list.append("'" + Cnt[k].replace("\\", "/") + "'")
            else:
                pth_list.append("''")

        # modify resources.py with the new paths
        strNew = "# # # start NiftyPET tools # # #\n"
        for i in range(len(key_list)):
            if pth_list[i] != "''":
                strNew += key_list[i] + " = " + pth_list[i] + "\n"
        rsrcNew = rsrc[:i0] + strNew + rsrc[i1:]
        f = open(resources_file, "w")
        f.write(rsrcNew)
        f.close()

    return Cnt



"""
This module provides information about CUDA resources (GPUs).
"""
from textwrap import dedent

from miutil import cuinfo


def dev_info(showprop=False):
    """
    Obtain information about installed GPU devices.
    """
    if showprop:
        for i in range(cuinfo.num_devices()):
            print(
                dedent(
                    """\
                    Name: {0:s}
                     mem: {1[0]:d} MiB total, {1[1]:d} MiB free, {1[2]:d} MiB used
                      CC: {2[0]:d}.{2[1]:d}\
                    """
                ).format(
                    cuinfo.name(i),
                    [mem >> 20 for mem in cuinfo.memory(i)],
                    cuinfo.compute_capability(i),
                )
            )
    return [
        (cuinfo.name(i), cuinfo.memory(i)[0] >> 20, *cuinfo.compute_capability(i))
        for i in range(cuinfo.num_devices())
    ]


gpuinfo = dev_info



# version detector. Precedence: installed dist, git, 'UNKNOWN'
try:
    from ._dist_ver import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version(root="../..", relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "UNKNOWN"



#!/usr/bin/env python3
"""initialise the NiftyPET ninst package"""
import platform

from tqdm.auto import tqdm

from . import cudasetup as cs
from . import install_tools as tls
from .install_tools import LOG_FORMAT

__all__ = [
    "LOG_FORMAT",
    "LogHandler",
    "path_resources",
    "resources",
    "dev_info",
    "gpuinfo",
]


class LogHandler(tls.LogHandler):
    """Custom formatting and tqdm-compatibility"""

    def handleError(self, record):
        super().handleError(record)
        raise IOError(record)

    def emit(self, record):
        """Write to tqdm's stream so as to not break progress-bars"""
        try:
            msg = self.format(record)
            tqdm.write(msg, file=self.stream, end=getattr(self, "terminator", "\n"))
            self.flush()
        except RecursionError:
            raise
        except Exception:
            self.handleError(record)


# log = logging.getLogger(__name__)
# technically bad practice to add handlers
# https://docs.python.org/3/howto/logging.html#library-config
# log.addHandler(LogHandler())  # do it anyway for convenience


path_resources = cs.path_niftypet_local()
resources = cs.get_resources()

if getattr(resources, "CC_ARCH", "") and platform.system() in ["Linux", "Windows"]:
    from .dinf import dev_info, gpuinfo
else:
    dev_info, gpuinfo = None, None



#!/usr/bin/env python
"""Tools for CUDA compilation and set-up for Python 3."""
import importlib
import logging
import os
import platform
import re
import shutil
import sys
from distutils.sysconfig import get_python_inc
from subprocess import PIPE, run
from textwrap import dedent

# from pkg_resources import resource_filename
try:
    from numpy import get_include as get_numpy_inc
except ImportError:
    pass
else:
    nphdr = get_numpy_inc()  # numpy header path

log = logging.getLogger(__name__)

prefix = sys.prefix
pyhdr = get_python_inc()  # Python header paths
minc_c = 3, 5  # minimum required CUDA compute capability
mincc = minc_c[0] * 10 + minc_c[1]


def path_niftypet_local():
    """Get the path to the local (home) folder for NiftyPET resources."""
    # if using conda put the resources in the folder with the environment name
    if "CONDA_DEFAULT_ENV" in os.environ:
        try:
            env = re.findall(r"envs[/\\](.*)[/\\]bin[/\\]python", sys.executable)[0]
        except IndexError:
            env = os.environ["CONDA_DEFAULT_ENV"]
        log.info("install> conda environment found: {}".format(env))
    else:
        env = ""
    # create the path for the resources files according to the OS platform
    if platform.system() in ("Linux", "Darwin"):
        path_resources = os.path.expanduser("~")
    elif platform.system() == "Windows":
        path_resources = os.getenv("LOCALAPPDATA")
    else:
        raise ValueError("Unknown operating system: {}".format(platform.system()))
    path_resources = os.path.join(path_resources, ".niftypet", env)

    return path_resources


def find_cuda():
    """Locate the CUDA environment on the system."""
    # search the PATH for NVCC
    for fldr in os.environ["PATH"].split(os.pathsep):
        cuda_path = os.path.join(fldr, "nvcc")
        if os.path.exists(cuda_path):
            cuda_path = os.path.dirname(os.path.dirname(cuda_path))
            break
        cuda_path = None

    if cuda_path is None:
        log.warning("nvcc compiler could not be found from the PATH!")
        return

    # serach for the CUDA library path
    lcuda_path = os.path.join(cuda_path, "lib64")
    if "LD_LIBRARY_PATH" in os.environ:
        if lcuda_path in os.environ["LD_LIBRARY_PATH"].split(os.pathsep):
            log.info("found CUDA lib64 in LD_LIBRARY_PATH: {}".format(lcuda_path))
    elif os.path.isdir(lcuda_path):
        log.info("found CUDA lib64 in: {}".format(lcuda_path))
    else:
        log.warning("folder for CUDA library (64-bit) could not be found!")

    return cuda_path, lcuda_path


def dev_setup():
    """figure out what GPU devices are available and choose the supported ones."""
    log.info(
        dedent(
            """
            --------------------------------------------------------------
            Setting up CUDA ...
            --------------------------------------------------------------"""
        )
    )
    # check first if NiftyPET was already installed and use the choice of GPU
    path_resources = path_niftypet_local()
    # if so, import the resources and get the constants
    if os.path.isfile(os.path.join(path_resources, "resources.py")):
        resources = get_resources()
    else:
        log.error("resources file not found/installed.")
        return

    # get all constants and check if device is already chosen
    Cnt = resources.get_setup()
    # if "CCARCH" in Cnt and "DEVID" in Cnt:
    #     log.info("using this CUDA architecture(s): {}".format(Cnt["CCARCH"]))
    #     return Cnt["CCARCH"]

    from miutil import cuinfo

    # map from CUDA device order (CC) to NVML order (PCI bus)
    nvml_id = [
        i
        for _, i in sorted(
            ((cuinfo.compute_capability(i), i) for i in range(cuinfo.num_devices())),
            reverse=True,
        )
    ]
    if "DEVID" in Cnt:
        devid = int(Cnt["DEVID"])
        ccstr = cuinfo.nvcc_flags(nvml_id[devid])
        ccs = ["{:d}{:d}".format(*cuinfo.compute_capability(nvml_id[devid]))]
    else:
        devid = 0
        devs = cuinfo.num_devices()
        if devs < 1:
            return ""
        ccstr = ";".join(
            sorted(
                {
                    cuinfo.nvcc_flags(i)
                    for i in range(devs)
                    if cuinfo.compute_capability(i) >= minc_c
                }
            )
        )
        if not ccstr:
            return ""
        ccs = sorted(
            {
                "{:d}{:d}".format(*cuinfo.compute_capability(i))
                for i in range(devs)
                if cuinfo.compute_capability(i) >= minc_c
            }
        )

    # passing this setting to resources.py
    fpth = os.path.join(
        path_resources, "resources.py"
    )  # resource_filename(__name__, 'resources/resources.py')
    with open(fpth, "r") as f:
        rsrc = f.read()
    # get the region of keeping in synch with Python
    i0 = rsrc.find("# # # start GPU properties # # #")
    i1 = rsrc.find("# # # end GPU properties # # #")
    # list of constants which will be kept in sych from Python
    cnt_dict = {"DEV_ID": str(devid), "CC_ARCH": repr(ccstr)}
    # update the resource.py file
    with open(fpth, "w") as f:
        f.write(rsrc[:i0])
        f.write("# # # start GPU properties # # #\n")
        for k, v in cnt_dict.items():
            f.write(k + " = " + v + "\n")
        f.write(rsrc[i1:])

    return ccs


def resources_setup(gpu=True):
    """
    This function checks CUDA devices, selects some and installs resources.py
    """
    log.info("installing file <resources.py> into home directory if it does not exist.")
    path_current = os.path.dirname(os.path.realpath(__file__))
    # get the path to the local resources.py (on Linux machines it is in ~/.niftypet)
    path_resources = path_niftypet_local()
    log.info("current path: {}".format(path_current))

    # does the local folder for niftypet exists? if not create one.
    if not os.path.exists(path_resources):
        os.makedirs(path_resources)
    # is resources.py in the folder?
    if not os.path.isfile(os.path.join(path_resources, "resources.py")):
        if os.path.isfile(os.path.join(path_current, "raw", "resources.py")):
            shutil.copyfile(
                os.path.join(path_current, "raw", "resources.py"),
                os.path.join(path_resources, "resources.py"),
            )
        else:
            raise IOError("could not find <resources.py")
    else:
        log.info(
            "<resources.py> already found in local NiftyPET folder: {}".format(
                path_resources
            )
        )
        get_resources()

    # find available GPU devices, select one or more and output the compilation flags
    # return gpuarch for cmake compilation
    return dev_setup() if gpu else ""


def get_resources(sys_append=True, reload=True):
    path_resources = path_niftypet_local()
    if sys_append:
        if path_resources not in sys.path:
            sys.path.append(path_resources)
    try:
        import resources
    except ImportError:
        log.error(
            dedent(
                """\
        --------------------------------------------------------------------------
        NiftyPET resources file <resources.py> could not be imported.
        It should be in ~/.niftypet/resources.py (Linux) or
        in //Users//USERNAME//AppData//Local//niftypet//resources.py (Windows)
        but likely it does not exists.
        --------------------------------------------------------------------------"""
            )
        )
        raise
    else:
        return importlib.reload(resources) if reload else resources


def cmake_cuda(
    path_source, path_build, nvcc_flags="", logfile_prefix="py_", msvc_version=""
):
    # CUDA installation
    log.info(
        dedent(
            """
            --------------------------------------------------------------
            CUDA compilation
            --------------------------------------------------------------"""
        )
    )

    if not os.path.isdir(path_build):
        os.makedirs(path_build)
    path_current = os.path.abspath(os.curdir)
    try:
        os.chdir(path_build)

        # cmake installation commands
        cmds = [
            [
                "cmake",
                path_source,
                f"-DPython3_ROOT_DIR={sys.prefix}",
                f"-DCUDA_NVCC_FLAGS={nvcc_flags}",
            ],
            ["cmake", "--build", "./"],
        ]

        if platform.system() == "Windows":
            cmds[0] += ["-G", msvc_version]
            cmds[1] += ["--config", "Release"]

        # run commands with logging
        cmakelogs = [
            "{logfile_prefix}cmake_config.log",
            "{logfile_prefix}cmake_build.log",
        ]
        errs = False
        for cmd, cmakelog in zip(cmds, cmakelogs):
            log.info("Command:%s", cmd)
            p = run(cmd, stdout=PIPE, stderr=PIPE)
            stdout = p.stdout.decode("utf-8")
            stderr = p.stderr.decode("utf-8")

            with open(cmakelog, "w") as fd:
                fd.write(stdout)

            if p.returncode:
                errs = True

            log.info(
                dedent(
                    """
                    ----------- compilation output ----------
                    %s
                    ------------------ end ------------------"""
                ),
                stdout,
            )

            if p.stderr:
                log.error(
                    dedent(
                        """
                        --------------- process errors ----------------
                        %s
                        --------------------- end ---------------------"""
                    ),
                    stderr,
                )
        if errs:
            raise SystemError("compilation failed")
    finally:
        os.chdir(path_current)



import logging

from setuptools import setup

from niftypet.ninst import cudasetup as cs
from niftypet.ninst import install_tools as tls

logging.basicConfig(level=logging.INFO, format=tls.LOG_FORMAT)


tls.check_platform()
ext = tls.check_depends()  # external dependencies

# install resources.py
if ext["cuda"] and ext["cmake"]:
    # select the supported GPU device and
    gpuarch = cs.resources_setup()
else:
    gpuarch = cs.resources_setup(gpu=False)

setup(use_scm_version=True)
